//****************************************************************************
//**
//**    SND_MAIN.CPP
//**    Sound Subsystem
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#include "sys_defs.h"
#include "snd_win.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SND_NUMPLAYSLOTS 32

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
// .wav info
typedef struct
{
	dword marker; // four-character chunk type marker
	dword length; // length of chunk after this header
} wavChunkHeader_t;

typedef struct
{
	wavChunkHeader_t header; // marker "fmt "
	word formatTag; // should be 1
	word numChannels; // 1=mono, 2=stereo
	dword samplesPerSecond; // playback rate
	dword averageBytesPerSecond; // generally samplesPerSecond*blockAlign
	word blockAlign; // block alignment, bytesPerSample = blockAlign / numChannels
	word formatSpecific; // reserved I think
} wavFormatChunk_t;

typedef struct
{
	wavChunkHeader_t header; // marker "data"
	byte data[1]; // data stream, variable-sized
} wavDataChunk_t;

typedef struct
{
	dword marker; // "RIFF"
	dword waveChunkLen; // size of wave data chunk that follows (i.e. usually filesize = waveChunkLen+8)
	dword waveMarker; // "WAVE"
	// every chunk that follows is up for grabs, but hunt down the "fmt " and "data" chunks
} wavFileHeader_t;

// play slots
typedef struct
{
	dword soundHandle;
	dword playHandle;
	dword priority;
} snd_playSlot_t;

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
snd_playSlot_t snd_playSlots[SND_NUMPLAYSLOTS];

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void* WaveResourceLoad(char* fileName, void* userData)
{
	dword fp;
	byte* fpBuf;
	wavFileHeader_t* fhdr;
	wavChunkHeader_t* chdr;
	wavDataChunk_t* data;
	wavFormatChunk_t* format;
	dword chunkPos;
	dword result;

	if (!(fp = XF_Open(fileName)))
		return(NULL);
	fpBuf = SYS_MALLOC(byte, XF_Size(fp));
	XF_Read(fpBuf, XF_Size(fp), fp);
	XF_Close(fp);
	
	fhdr = (wavFileHeader_t*)fpBuf;
	if ((fhdr->marker != ('R' + ('I'<<8) + ('F'<<16) + ('F'<<24)))
	 || (fhdr->waveMarker != ('W' + ('A'<<8) + ('V'<<16) + ('E'<<24))))
	{
		return(NULL);
		SYS_FREE(fpBuf);
	}
	data = NULL;
	format = NULL;
	for (chunkPos=sizeof(wavFileHeader_t);
	     chunkPos<(fhdr->waveChunkLen+sizeof(wavChunkHeader_t));
		 chunkPos+=(chdr->length+sizeof(wavChunkHeader_t)) )
	{
		if ((data) && (format))
			break;
		chdr = (wavChunkHeader_t*)(fpBuf+chunkPos);
		if (chdr->marker == ('f' + ('m'<<8) + ('t'<<16) + (' '<<24)))
			format = (wavFormatChunk_t*)chdr;
		else if (chdr->marker == ('d' + ('a'<<8) + ('t'<<16) + ('a'<<24)))
			data = (wavDataChunk_t*)chdr;
	}
	if ((!data) || (!format))
	{
		SYS_FREE(fpBuf);
		return(NULL);
	}

	result = SNDW_LoadSound(format->samplesPerSecond, (format->blockAlign / format->numChannels)*8,
		format->numChannels, data->header.length, data->data, (dword)userData);
	SYS_FREE(fpBuf);
	return((void*)result);
}

static void WaveResourceFree(res_id_t id)
{
	dword soundHandle = (dword)RES_PtrForId(id, NULL);
	SNDW_FreeSound(soundHandle);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// init/shutdown

void SND_Init()
{
	SNDW_Init();
}

void SND_Shutdown()
{
	SNDW_Shutdown();
}

// sound control

res_id_t SND_RegisterSound(char* fileName, dword maxPlayHandles)
{
	return(RES_Register(fileName, RES_CACHE_LEVEL, WaveResourceLoad, WaveResourceFree, (void*)maxPlayHandles));
}

// play control

dword SND_PlaySound(res_id_t soundId, dword priority, float volume,
					float freq, float pan, bbool looping, bbool allowDuplicate)
{
	dword i, soundHandle, playHandle;
	
	if (!(soundHandle = (dword)RES_PtrForId(soundId, NULL)))
		return(0);

	// check if this sound is already playing, if so then don't duplicate it
    if (!allowDuplicate)
    {
	    for (i=0;i<SND_NUMPLAYSLOTS;i++)
	    {
		    if ((snd_playSlots[i].soundHandle == soundHandle)
		     && (snd_playSlots[i].playHandle)
		     && (SNDW_IsSoundPlaying(snd_playSlots[i].playHandle)))
			    return(i+1);
	    }
    }

	// find an available slot
	for (i=0;i<SND_NUMPLAYSLOTS;i++)
	{
		if (!snd_playSlots[i].playHandle)
			break;
		if (!SNDW_IsSoundPlaying(snd_playSlots[i].playHandle))
			break;
	}
	if (i==SND_NUMPLAYSLOTS)
	{
		// no slots available, find one with lower priority or "low" priority, whichever comes first
		for (i=0;i<SND_NUMPLAYSLOTS;i++)
		{
			if ((!snd_playSlots[i].priority) || (snd_playSlots[i].priority < priority))
			{
				SNDW_StopSound(snd_playSlots[i].playHandle);
				break;
			}
		}
	}
	if (i==SND_NUMPLAYSLOTS)
	{
		// still no slots available, forget it
		return(0);
	}
	
	// get a play handle
	if (!(playHandle = SNDW_PlaySound(soundHandle, looping)))
		return(0);
	snd_playSlots[i].soundHandle = soundHandle;
	snd_playSlots[i].playHandle = playHandle;
	snd_playSlots[i].priority = priority;
	SNDW_SetSoundVolume(playHandle, volume);
	SNDW_SetSoundPan(playHandle, pan);
	SNDW_SetSoundFrequency(playHandle, freq);
	return(i+1);
}

bbool SND_StopSound(dword slot)
{
	if ((!slot) || (slot > SND_NUMPLAYSLOTS))
		return(0);
	if (!snd_playSlots[slot-1].playHandle)
		return(0);
	SNDW_StopSound(snd_playSlots[slot-1].playHandle);
	return(1);
}

bbool SND_ChangeSoundVolume(dword slot, float volume)
{
	if ((!slot) || (slot > SND_NUMPLAYSLOTS))
		return(0);
	if (!snd_playSlots[slot-1].playHandle)
		return(0);
	SNDW_SetSoundVolume(snd_playSlots[slot-1].playHandle, volume);
	return(1);
}

bbool SND_ChangeSoundFrequency(dword slot, float freq)
{
	if ((!slot) || (slot > SND_NUMPLAYSLOTS))
		return(0);
	if (!snd_playSlots[slot-1].playHandle)
		return(0);
	SNDW_SetSoundFrequency(snd_playSlots[slot-1].playHandle, freq);
	return(1);
}

bbool SND_ChangeSoundPan(dword slot, float pan)
{
	if ((!slot) || (slot > SND_NUMPLAYSLOTS))
		return(0);
	if (!snd_playSlots[slot-1].playHandle)
		return(0);
	SNDW_SetSoundPan(snd_playSlots[slot-1].playHandle, pan);
	return(1);
}

// misc operations

void SND_SetVolume(float volume)
{
	SNDW_SetGlobalVolume(volume);
}

float SND_GetVolume()
{
	return(SNDW_GetGlobalVolume());
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SND_MAIN.CPP
//**
//****************************************************************************

